import logging
from typing import Optional

from pydantic import StrictBool, StrictStr, validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.gcp_gcs import GCPGCSConnOp

logger = logging.getLogger(__name__)


class GCPGCSDeleteKeyOpConfigModel(BaseOpConfigModel):
    gcs_conn_id: StrictStr
    bucket: StrictStr
    key: StrictStr
    check_if_exists: StrictBool

    @validator("key", pre=False)
    def validate_key(cls, key: str):
        if key.endswith("/"):
            raise ValueError(f"{key=} must not end with a '/'")

        if key.startswith("/"):
            raise ValueError(f"{key=} must not start with a '/'")

        return key

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSDeleteKeyOpMetadataModel(BaseOpMetadataModel):
    bucket: StrictStr
    key: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSDeleteKeyOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSDeleteKeyOp(BaseConsumerOp):
    OP_TYPE = "consumer.gcp.gcs.delete_key"
    OP_CONFIG_MODEL = GCPGCSDeleteKeyOpConfigModel
    OP_METADATA_MODEL = GCPGCSDeleteKeyOpMetadataModel
    OP_AUDIT_MODEL = GCPGCSDeleteKeyOpAuditModel

    templated_fields = None

    def __init__(
        self,
        gcs_conn_id: str,
        bucket: str,
        key: str,
        check_if_exists: bool,
        **kwargs,
    ):
        """."""
        self.gcs_conn_id = gcs_conn_id
        self.bucket = bucket
        self.key = key
        self.check_if_exists = check_if_exists

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.gcs_conn: GCPGCSConnOp = op_manager.get_connector(op_id=self.gcs_conn_id)

    def run(self, time_step: TimeStep, msg: Optional[OpMsg] = None) -> OpMsg:
        """."""
        logger.info(f"GCPGCSDeleteKeyOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="GCPGCSDeleteKeyOp.run:"
        )

        # First assume, that the object exists. This means we will delete the key.
        # Later, if we find out that the key does not exist, we will set this to
        # False. If it is False, then we will not attempt to delete.
        is_key_exists = True

        if self.check_if_exists:
            logger.info(
                f"GCPGCSDeleteKeyOp.run: {self.check_if_exists=}. Checking to see "
                f"if key exists."
            )
            is_key_exists = self.gcs_conn.is_key_exists(
                bucket=self.bucket, key=self.key
            )
            logger.info(
                f"GCPGCSDeleteKeyOp.run: Done checking to see if key exists. "
                f"{is_key_exists=}."
            )

        if is_key_exists:
            logger.info(
                f"GCPGCSDeleteKeyOp.run: {is_key_exists=}. Deleting key "
                f"{self.bucket} | {self.key=}"
            )
            self.gcs_conn.delete_key(bucket=self.bucket, key=self.key)
            logger.info(
                f"GCPGCSDeleteKeyOp.run: Done deleting key {self.bucket} | {self.key=}"
            )
        else:
            logger.info(
                f"GCPGCSDeleteKeyOp.run: {is_key_exists=}. Since key does not exist "
                f"we will not attempt to delete it."
            )

        return OpMsg(
            data=None,
            metadata=GCPGCSDeleteKeyOpMetadataModel(
                bucket=self.bucket,
                key=self.key,
            ),
            audit=GCPGCSDeleteKeyOpAuditModel(),
        )



import logging
import os
import subprocess
import time
from typing import List, Optional

from pydantic import StrictBool, StrictInt, StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class ShellRunConsumerOpConfigModel(BaseOpConfigModel):
    cmd: StrictStr
    accepted_exit_codes: Optional[conlist(StrictInt, min_items=1)] = None
    log_cmd: Optional[StrictBool] = False
    log_stdout: Optional[StrictBool] = False

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ShellRunConsumerOpMetadataModel(BaseOpMetadataModel):
    exit_code: int

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ShellRunConsumerOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ShellRunConsumerOp(BaseConsumerOp):
    OP_TYPE = "consumer.shell.run"
    OP_CONFIG_MODEL = ShellRunConsumerOpConfigModel
    OP_METADATA_MODEL = ShellRunConsumerOpMetadataModel
    OP_AUDIT_MODEL = ShellRunConsumerOpAuditModel

    templated_fields = None

    def __init__(
        self,
        cmd: str,
        accepted_exit_codes: Optional[List[int]] = None,
        log_cmd: bool = False,
        log_stdout: bool = False,
        **kwargs,
    ):
        self.cmd = cmd
        self.accepted_exit_codes = (
            [0] if accepted_exit_codes is None else accepted_exit_codes
        )
        self.log_cmd = log_cmd
        self.log_stdout = log_stdout

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self, time_step: TimeStep, msg: Optional[OpMsg] = None) -> OpMsg:
        """."""
        logger.info(f"ShellRunConsumerOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="ShellRunConsumerOp.run:"
        )

        if self.log_cmd:
            logger.info(f"ShellRunConsumerOp.run: Running command | {self.cmd=}")
        else:
            logger.info(f"ShellRunConsumerOp.run: Running command | not logging cmd")

        start = time.time()

        proc = subprocess.Popen(
            self.cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            env=os.environ.copy(),
        )

        if not self.log_stdout:
            logger.info(
                f"ShellRunConsumerOp.run: {self.log_stdout=}. Not logging stdout. " f""
            )

        # https://gist.github.com/JLeClerc/831d400763b7020599d9
        for line in iter(proc.stdout.readline, b""):
            if self.log_stdout:
                logger.info(
                    f"ShellRunConsumerOp.run: STDOUT | {line.decode('utf-8').rstrip()}"
                )
            else:
                logger.info(
                    f"ShellRunConsumerOp.run: {self.log_stdout=} | Still running ..."
                )

        exit_code = proc.wait()

        logger.info(
            f"ShellRunConsumerOp.run: Finished running command with {exit_code=}. "
            f"Time taken: {time.time() - start:0.2f} seconds."
        )

        if exit_code not in self.accepted_exit_codes:
            raise ValueError(
                f"ShellRunConsumerOp.run: Command exited with {exit_code=} which is "
                f"not in the list of {self.accepted_exit_codes=}."
            )

        return OpMsg(
            data=None,
            metadata=ShellRunConsumerOpMetadataModel(exit_code=exit_code),
            audit=ShellRunConsumerOpAuditModel(),
        )



import logging
from typing import Any, Dict, Literal, Optional

import pangres as pg
from pydantic import BaseModel, StrictBool, StrictInt, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.postgres import PostgresConnOp
from nr_ops.utils.sqlalchemy.dtype_lookup import lookup_sqlalchemy_dtype

logger = logging.getLogger(__name__)


class UpsertConfigModel(BaseModel):
    # Reference: https://github.com/ThibTrip/pangres/wiki/Upsert
    table_name: StrictStr
    schema_: StrictStr
    if_row_exists: Literal["ignore", "update"]
    dtype: Optional[Dict[StrictStr, StrictStr]] = None
    create_schema: StrictBool = False
    create_table: StrictBool = False
    add_new_columns: StrictBool = False
    chunksize: Optional[StrictInt] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False
        fields = {"schema_": "schema"}

    # Source: https://github.com/pydantic/pydantic/issues/153#issuecomment-379695371
    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        d["schema"] = d.pop("schema_")
        return d


class PangresDFToSQLDBOpOpConfigModel(BaseOpConfigModel):
    postgres_conn_id: StrictStr
    upsert_config: UpsertConfigModel

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PangresDFToSQLDBOpOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PangresDFToSQLDBOpOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PangresDFToSQLDBOp(BaseConsumerOp):
    OP_TYPE = "consumer.pangres.df_to_sql_db"
    OP_CONFIG_MODEL = PangresDFToSQLDBOpOpConfigModel
    OP_METADATA_MODEL = PangresDFToSQLDBOpOpMetadataModel
    OP_AUDIT_MODEL = PangresDFToSQLDBOpOpAuditModel

    templated_fields = None

    def __init__(self, postgres_conn_id: str, upsert_config: Dict[str, Any], **kwargs):
        """."""
        self.postgres_conn_id = postgres_conn_id
        self.upsert_config = upsert_config
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.postgres: PostgresConnOp = op_manager.get_connector(
            op_id=self.postgres_conn_id
        )

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info("PangresDFToSQLDBOpOp: Running.")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="PangresDFToSQLDBOpOp.run:"
        )

        if self.upsert_config is None:
            col_to_dtype_obj_map = None
        else:
            col_to_dtype_obj_map = {
                column_name: lookup_sqlalchemy_dtype(dtype)
                for column_name, dtype in self.upsert_config.get("dtype", {}).items()
            }

        pg.upsert(
            con=self.postgres.get_engine(),
            df=msg.data,
            create_schema=self.upsert_config.get("create_schema", False),
            create_table=self.upsert_config.get("create_table", False),
            add_new_columns=self.upsert_config.get("add_new_columns", False),
            adapt_dtype_of_empty_db_columns=False,
            chunksize=self.upsert_config.get("chunksize", None),
            yield_chunks=False,
            table_name=self.upsert_config.get("table_name"),
            schema=self.upsert_config.get("schema"),
            if_row_exists=self.upsert_config.get("if_row_exists"),
            dtype=col_to_dtype_obj_map,
        )

        return OpMsg(
            data=None,
            metadata=PangresDFToSQLDBOpOpMetadataModel(),
            audit=PangresDFToSQLDBOpOpAuditModel(),
        )



import logging
from typing import Generator

import pandas as pd

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpTimeStepMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseOpConfigModel, BaseTimeStepOp

logger = logging.getLogger(__name__)


class MockTimeStepOpConfigModel(BaseOpConfigModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MockTimeStepOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MockTimeStepOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class MockTimeStepOp(BaseTimeStepOp):
    OP_TYPE = "ts.mock"
    OP_CONFIG_MODEL = MockTimeStepOpConfigModel
    OP_METADATA_MODEL = MockTimeStepOpMetadataModel
    OP_AUDIT_MODEL = MockTimeStepOpAuditModel

    templated_fields = None

    def __init__(self, **kwargs):
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self) -> Generator[OpTimeStepMsg, None, None]:
        """."""
        logger.info(f"MockTimeStepOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=None, msg=None, log_prefix="MockTimeStepOp.run:")

        time_step = TimeStep(
            start=pd.Timestamp.now(tz="UTC"),
            end=pd.Timestamp.now(tz="UTC") + pd.Timedelta(days=1),
            tz="UTC",
        )
        yield OpTimeStepMsg(
            data=time_step,
            metadata=MockTimeStepOpMetadataModel(),
            audit=MockTimeStepOpAuditModel(),
        )



import itertools
import logging
import time
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple

import pandas as pd
from pydantic import Field, StrictBool, StrictInt, StrictStr, conlist, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp
from nr_ops.ops.ops.generator_ops.airflow.dagruns.get_dagrun import (
    AirflowDagRunGetDagRunOp,
)
from nr_ops.ops.ops.generator_ops.airflow.dagruns.trigger_dagrun import (
    AirflowDagRunTriggerDagRunOp,
    AirflowDagRunTriggerDagRunOpConfigModel,
)

logger = logging.getLogger(__name__)


class AirflowDagRunTriggerDagAndWaitUntilCompletionOpConfigModel(BaseOpConfigModel):
    trigger_dagrun_config: AirflowDagRunTriggerDagRunOpConfigModel
    trigger_accepted_status_codes: conlist(StrictInt, min_items=1) = Field(
        default_factory=lambda: [200]
    )
    poll_interval: float

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowDagRunTriggerDagAndWaitUntilCompletionOpMetadataModel(BaseOpMetadataModel):
    etl_request_start_ts: StrictStr
    etl_response_end_ts: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowDagRunTriggerDagAndWaitUntilCompletionOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowDagRunTriggerDagAndWaitUntilCompletionOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.airflow.dagruns.trigger_dagrun_and_wait_until_completion"
    OP_CONFIG_MODEL = AirflowDagRunTriggerDagAndWaitUntilCompletionOpConfigModel
    OP_METADATA_MODEL = AirflowDagRunTriggerDagAndWaitUntilCompletionOpMetadataModel
    OP_AUDIT_MODEL = AirflowDagRunTriggerDagAndWaitUntilCompletionOpAuditModel

    templated_fields = None

    def __init__(
        self,
        trigger_dagrun_config: Dict[str, Any],
        trigger_accepted_status_codes: bool,
        poll_interval: float,
        **kwargs,
    ):
        """.

        Args:
            trigger_accepted_status_codes (bool): The status codes that are accepted
                from the trigger dagrun op. If the trigger dagrun op returns a status
                code that is not in this list, an exception is raised.

                Usually, the trigger dagrun op returns 200 (success). Sometimes it
                returns 409 (conflict) if the dagrun already exists. In this case, it
                is still accepted and the dagrun is waited for.
        """
        self.trigger_dagrun_config = trigger_dagrun_config
        self.trigger_accepted_status_codes = trigger_accepted_status_codes
        self.poll_interval = poll_interval

        self.templated_fields = kwargs.get("templated_fields", [])

        self.trigger_dagrun_config_model = AirflowDagRunTriggerDagRunOpConfigModel(
            **trigger_dagrun_config
        )

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(
            op_id=self.trigger_dagrun_config_model.http_conn_id
        )

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run:",
        )

        logger.info(
            f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: Initializing and "
            f"running AirflowDagRunTriggerDagRunOp."
        )

        trigger_dagrun_op = AirflowDagRunTriggerDagRunOp(**self.trigger_dagrun_config)
        trigger_dagrun_op_msg: OpMsg = list(trigger_dagrun_op.run(time_step=time_step))[
            0
        ]

        logger.info(
            f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: Done running "
            f"AirflowDagRunTriggerDagRunOp, "
            f"returned {trigger_dagrun_op_msg.data['status_code']=}."
        )

        if (
            trigger_dagrun_op_msg.data["status_code"]
            not in self.trigger_accepted_status_codes
        ):
            raise Exception(
                f"Trigger dagrun did not return one of "
                f"{self.trigger_accepted_status_codes}. "
                f"{trigger_dagrun_op_msg.data['status_code']=} and "
                f"{trigger_dagrun_op_msg.data=}"
            )

        # Depending on the trigger status code, we need to get the dag_run_id from
        # different places.
        if trigger_dagrun_op_msg.data["status_code"] == 200:
            dag_run_id = trigger_dagrun_op_msg.data["response"]["dag_run_id"]
        elif trigger_dagrun_op_msg.data["status_code"] == 409:
            dag_run_id = self.trigger_dagrun_config_model.dag_run_id
            if not dag_run_id:
                raise Exception(
                    f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp: "
                    f"Since trigger_dagrun_op_msg.data['status_code']==409, the "
                    f"dag_run_id is not returned in the response object so we have to "
                    f"get it from the trigger_dagrun_config_model. However, the model "
                    f"is not guaranteed to have dag_run_id set (i.e. the config "
                    f"provided to this op does have dag_run_id set). This situation of "
                    f"the model not containing dag_run_id has just occurred. "
                    f"You will need to find a workaround for this situation."
                )
        else:
            raise NotImplementedError()

        logger.info(
            f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: Initializing "
            f"AirflowDagRunGetDagRunOp."
        )

        get_dagrun_op = AirflowDagRunGetDagRunOp(
            http_conn_id=self.trigger_dagrun_config_model.http_conn_id,
            dag_id=self.trigger_dagrun_config_model.dag_id,
            dag_run_id=dag_run_id,
            # Accept any status code since status codes are handled within this op.
            accepted_status_codes=None,
        )

        logger.info(
            f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: Successfully "
            f"initialized AirflowDagRunGetDagRunOp."
        )

        for wait_iteration in itertools.count(1):
            logger.info(
                f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: Running "
                f"{wait_iteration=} (1-indexed). Starting off with waiting for "
                f"{self.poll_interval=} seconds."
            )

            time.sleep(self.poll_interval)

            logger.info(
                f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: Running "
                f"AirflowDagRunGetDagRunOp for {wait_iteration=} (1-indexed)."
            )

            get_dagrun_op_msg: OpMsg = list(get_dagrun_op.run(time_step=time_step))[0]

            if get_dagrun_op_msg.data["status_code"] != 200:
                raise Exception(
                    f"Get dagrun did not return 200. "
                    f"{get_dagrun_op_msg.data['status_code']=} and "
                    f"{get_dagrun_op_msg.data=}"
                )

            if get_dagrun_op_msg.data["response"]["state"] in ["queued", "running"]:
                logger.info(
                    f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: "
                    f"AirflowDagRunGetDagRunOp for {wait_iteration=} (1-indexed) "
                    f"returned {get_dagrun_op_msg.data['response']['state']=}. "
                )
                continue

            elif get_dagrun_op_msg.data["response"]["state"] in ["success", "failed"]:
                logger.info(
                    f"AirflowDagRunTriggerDagAndWaitUntilCompletionOp.run: "
                    f"AirflowDagRunGetDagRunOp for {wait_iteration=} (1-indexed) "
                    f"returned {get_dagrun_op_msg.data['response']['state']=}. "
                    f"DAGRun has completed. Yielding get_dagrun_op_msg."
                )
                yield get_dagrun_op_msg
                # Must return otherwise the generator will continue to infinity.
                return

            else:
                raise NotImplementedError(
                    f"Unrecognized state {get_dagrun_op_msg.data['response']['state']=}"
                )



from typing import Optional

import pandas as pd
from pydantic import BaseModel, Field, StrictStr, conint


class MetadataModel(BaseModel):
    created_at: StrictStr = Field(
        default_factory=lambda: pd.Timestamp.now(tz="UTC").isoformat()
    )

    class Config:
        # NOTE!!! This is required to allow NEW fields to be added to the model
        # by the user. This is useful for adding metadata to the model.
        extra = "allow"
        # NOTE!!! We don't allow any arbitrary types in the model.
        arbitrary_types_allowed = False


class TimeStep(BaseModel):
    index: Optional[conint(strict=True, ge=0)] = None
    start: pd.Timestamp
    end: pd.Timestamp
    tz: StrictStr
    metadata: MetadataModel = Field(default_factory=MetadataModel)

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = True

    def __repr__(self):
        """Overwrite the default __repr__ with something more human-readable.

        This is useful when logging an initialized object.
        """
        return (
            f"TimeStep: "
            f"index={self.index} | "
            f"start='{self.start.isoformat()}' | "
            f"end='{self.end.isoformat()}' | "
            f"tz='{self.tz}'"
            # json.dumps also handles None and converts it to "null"
            f"metadata='{self.metadata.json()}'"
        )

    def to_json_dict(self) -> dict:
        return {
            "index": self.index,
            "start": self.start.isoformat(),
            "end": self.end.isoformat(),
            "tz": self.tz,
            "metadata": self.metadata.dict(),
        }

    @classmethod
    def mock_init(cls):
        return cls(
            index=0,
            start=pd.Timestamp(0, tz="UTC"),
            end=pd.Timestamp(0, tz="UTC") + pd.Timedelta(days=1),
            tz="UTC",
            metadata=MetadataModel(),
        )



"""."""
import logging
from typing import Literal

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class PythonFileHookConnOpConfigModel(BaseOpConfigModel):
    path: StrictStr
    mode: Literal[
        "r",
        "r+",
        "w",
        "w+",
        "a",
        "a+",
        "x",
        "rb",
        "rb+",
        "wb",
        "wb+",
        "ab",
        "ab+",
        "xb",
    ]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonFileHookConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonFileHookConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonFileHookConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.python.file"
    OP_CONFIG_MODEL = PythonFileHookConnOpConfigModel
    OP_METADATA_MODEL = PythonFileHookConnOpMetadataModel
    OP_AUDIT_MODEL = PythonFileHookConnOpAuditModel

    templated_fields = None

    def __init__(
        self,
        path: str,
        mode: Literal[
            "r",
            "r+",
            "w",
            "w+",
            "a",
            "a+",
            "x",
            "rb",
            "rb+",
            "wb",
            "wb+",
            "ab",
            "ab+",
            "xb",
        ],
        **kwargs,
    ):
        self.path = path
        self.mode = mode
        self.templated_fields = kwargs.get("templated_fields", [])

        self._file = open(self.path, self.mode)

    def run(self) -> OpMsg:
        """."""
        logger.info(f"PythonFileHookConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="PythonFileHookConnOp.run:"
        )

        return OpMsg(
            data=self,
            metadata=PythonFileHookConnOpMetadataModel(),
            audit=PythonFileHookConnOpAuditModel(),
        )

    def get_reference(self):
        """."""
        return self._file

    def get_read(self):
        """."""
        self._file.read()

    def put_write(self, item):
        """."""
        self._file.write(item)

    def put_write_line(self, item):
        """."""
        self._file.write(item)
        self._file.write("\n")

    def put_write_and_flush(self, item):
        """."""
        self._file.write(item)
        self._file.flush()

    def put_write_line_and_flush(self, item):
        """."""
        self._file.write(item)
        self._file.write("\n")
        self._file.flush()

    # TODO: Seek



import logging
import time
from typing import Generator, Optional

from pydantic import conint

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class SleepGeneratorOpConfigModel(BaseOpConfigModel):
    seconds: conint(ge=0) = 0

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SleepGeneratorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SleepGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SleepGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.sleep"
    OP_CONFIG_MODEL = SleepGeneratorOpConfigModel
    OP_METADATA_MODEL = SleepGeneratorOpMetadataModel
    OP_AUDIT_MODEL = SleepGeneratorOpAuditModel

    templated_fields = None

    def __init__(self, seconds: int, **kwargs):
        self.seconds = seconds
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"SleepGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="SleepGeneratorOp.run:"
        )

        logger.info(f"SleepGeneratorOp.run: Sleeping for {self.seconds=}")
        time.sleep(self.seconds)
        logger.info(
            f"SleepGeneratorOp.run: Done sleeping for {self.seconds=}. "
            f"Yielding input OpMsg as is."
        )

        yield OpMsg(
            data=msg.data,
            metadata=SleepGeneratorOpMetadataModel(),
            audit=SleepGeneratorOpAuditModel(),
        )



import logging
from typing import Any, Dict, Generator, List, Optional

import pandas as pd
from pydantic import StrictStr, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpTimeStepMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseOpConfigModel, BaseTimeStepOp

logger = logging.getLogger(__name__)


class SimpleStartOffsetTimeStepOpConfigModel(BaseOpConfigModel):
    start: StrictStr
    end_offset: StrictStr
    tz: StrictStr
    metadata: Optional[Dict[str, Any]] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SimpleStartOffsetTimeStepOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SimpleStartOffsetTimeStepOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SimpleStartOffsetTimeStepOp(BaseTimeStepOp):
    OP_TYPE = "ts.simple_start_offset"
    OP_CONFIG_MODEL = SimpleStartOffsetTimeStepOpConfigModel
    OP_METADATA_MODEL = SimpleStartOffsetTimeStepOpMetadataModel
    OP_AUDIT_MODEL = SimpleStartOffsetTimeStepOpAuditModel

    templated_fields = None

    def __init__(
        self,
        start: str,
        end_offset: str,
        tz: str,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.start = start
        self.end_offset = end_offset
        self.tz = tz
        self.metadata = {} if metadata is None else metadata

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(self) -> Generator[OpTimeStepMsg, None, None]:
        """."""
        logger.info(f"SimpleStartOffsetTimeStepOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="SimpleStartOffsetTimeStepOp.run:"
        )

        end = pd.Timestamp(self.start, tz=self.tz) + pd.Timedelta(self.end_offset)

        time_step = TimeStep(
            start=self.start,
            end=end,
            tz=self.tz,
            metadata=self.metadata,
        )

        logger.info(f"SimpleStartOffsetTimeStepOp.run: {time_step=}")

        yield OpTimeStepMsg(
            data=time_step,
            metadata=SimpleStartOffsetTimeStepOpMetadataModel(),
            audit=SimpleStartOffsetTimeStepOpAuditModel(),
        )



import logging
from io import StringIO
from typing import Any, Generator, Literal, Optional

import numpy as np
import pandas as pd
from pydantic import StrictBool, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.utils.eval.eval_globals import EVAL_GLOBALS

logger = logging.getLogger(__name__)


class EvalExprConditionalOpConfigModel(BaseOpConfigModel):
    conditional_expr: StrictStr
    on_true_behavior: Literal["yield_input", "yield_output", "consume"]
    on_false_behavior: Literal["yield_input", "yield_output", "consume"]
    expr: StrictStr
    msg_var_name: Optional[StrictStr] = None
    time_step_var_name: Optional[StrictStr] = None
    op_manager_var_name: Optional[StrictStr] = None
    log_output: StrictBool = False
    iterate_over_output: StrictBool = False

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprConditionalOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprConditionalOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class EvalExprConditionalOp(BaseGeneratorOp):
    """Store output of eval_expr as metadata. Return input data as is."""

    OP_TYPE = "generator.eval_expr_conditional"
    OP_CONFIG_MODEL = EvalExprConditionalOpConfigModel
    OP_METADATA_MODEL = EvalExprConditionalOpMetadataModel
    OP_AUDIT_MODEL = EvalExprConditionalOpAuditModel

    templated_fields = None

    def __init__(
        self,
        conditional_expr: str,
        on_true_behavior: str,
        on_false_behavior: str,
        expr: str,
        msg_var_name: Optional[str] = None,
        time_step_var_name: Optional[str] = None,
        op_manager_var_name: Optional[str] = None,
        log_output: bool = False,
        iterate_over_output: bool = False,
        **kwargs,
    ):
        """."""
        self.conditional_expr = conditional_expr
        self.on_true_behavior = on_true_behavior
        self.on_false_behavior = on_false_behavior
        self.expr = expr
        self.msg_var_name = "msg" if msg_var_name is None else msg_var_name
        self.time_step_var_name = (
            "time_step" if time_step_var_name is None else time_step_var_name
        )
        self.op_manager_var_name = (
            "op_manager" if op_manager_var_name is None else op_manager_var_name
        )
        self.log_output = log_output
        self.iterate_over_output = iterate_over_output

        self.templated_fields = kwargs.get("templated_fields", [])

        self.op_manager = get_global_op_manager()

    def eval_output(self, time_step: TimeStep, msg: OpMsg):
        """."""
        output = eval(
            self.expr,
            {
                **EVAL_GLOBALS,
                self.msg_var_name: msg,
                self.time_step_var_name: time_step,
                self.op_manager_var_name: self.op_manager,
            },
        )

        logger.info(
            f"EvalExprConditionalOp.run: Evaluated output expression: "
            f"{type(output)=}"
        )

        # Special logging for pandas DataFrames
        if isinstance(output, pd.DataFrame):
            buf = StringIO()
            output.info(buf=buf)
            buf.seek(0)
            logger.info(f"EvalExprConditionalOp.run: DataFrame info:\n{buf.read()}")

        if self.log_output:
            logger.info(f"EvalExprConditionalOp.run: Logging output:\n{output=}")

        return output

    def behavior_consume(self):
        """."""
        logger.info(
            f"EvalExprConditionalOp.run: "
            f"Consuming and returning i.e. stopping the generator."
        )
        return

    def behavior_yield_input(self, input_msg: OpMsg):
        """."""
        logger.info(f"EvalExprConditionalOp.run: Yielding input msg")
        yield OpMsg(
            data=input_msg.data,
            metadata=EvalExprConditionalOpMetadataModel(),
            audit=EvalExprConditionalOpAuditModel(),
        )

    def behavior_yield_output(self, output: Any):
        """."""
        if self.iterate_over_output:
            logger.info(
                f"EvalExprConditionalOp.run: Iterating over output of "
                f"{type(output)=} and yielding items"
            )

            for item_i, item in enumerate(output):
                logger.info(
                    f"EvalExprConditionalOp.run: "
                    f"Iterating over output of {type(output)=}, "
                    f"yielding {item_i=} of {type(item)=}"
                )
                yield OpMsg(
                    data=item,
                    metadata=EvalExprConditionalOpMetadataModel(),
                    audit=EvalExprConditionalOpAuditModel(),
                )
        else:
            logger.info(
                f"EvalExprConditionalOp.run: Yielding output of {type(output)=}"
            )
            yield OpMsg(
                data=output,
                metadata=EvalExprConditionalOpMetadataModel(),
                audit=EvalExprConditionalOpAuditModel(),
            )

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"EvalExprConditionalOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="EvalExprConditionalOp.run:"
        )

        logger.info(
            f"EvalExprOp.run: Evaluating conditional expression: "
            f"{self.conditional_expr=}"
        )

        conditional_output = eval(
            self.conditional_expr,
            {
                **EVAL_GLOBALS,
                self.msg_var_name: msg,
                self.time_step_var_name: time_step,
                self.op_manager_var_name: self.op_manager,
            },
        )

        logger.info(
            f"EvalExprConditionalOp.run: Evaluated conditional expression: "
            f"{type(conditional_output)=}"
        )

        if not isinstance(conditional_output, (bool, np.bool_)):
            raise TypeError(
                f"EvalExprConditionalOp.run: Conditional expression must evaluate to "
                f"bool or np.bool_. Got {type(conditional_output)=}"
            )

        if isinstance(conditional_output, np.bool_):
            conditional_output = bool(conditional_output)

        logger.info(
            f"EvalExprConditionalOp.run: Evaluated conditional expression: "
            f"{conditional_output=}"
        )

        if conditional_output is True:
            logger.info(
                f"EvalExprConditionalOp.run: {conditional_output=}. "
                f"Executing {self.on_true_behavior=}."
            )
            if self.on_true_behavior == "consume":
                return self.behavior_consume()
            elif self.on_true_behavior == "yield_input":
                yield from self.behavior_yield_input(msg)
            elif self.on_true_behavior == "yield_output":
                output = self.eval_output(time_step, msg)
                yield from self.behavior_yield_output(output)
            else:
                raise NotImplementedError()
        else:
            logger.info(
                f"EvalExprConditionalOp.run: {conditional_output=}. "
                f"Executing {self.on_false_behavior=}."
            )
            if self.on_false_behavior == "consume":
                return self.behavior_consume()
            elif self.on_false_behavior == "yield_input":
                yield from self.behavior_yield_input(msg)
            elif self.on_false_behavior == "yield_output":
                output = self.eval_output(time_step, msg)
                yield from self.behavior_yield_output(output)
            else:
                raise NotImplementedError()



import logging
from typing import Optional

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.google_ads import GoogleAdsConnectorOp

logger = logging.getLogger(__name__)


class GoogleAdsUploadOfflineConversionOpConfigModel(BaseOpConfigModel):
    google_ads_conn_id: StrictStr
    customer_id: StrictStr
    conversion_action_id: StrictStr
    gclid: StrictStr
    conversion_date_time: StrictStr
    conversion_value: StrictStr
    currency_code: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsUploadOfflineConversionOpMetadataModel(BaseOpMetadataModel):
    successful_upload: bool
    gclid: str
    conversion_date_time: str
    conversion_value: str

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsUploadOfflineConversionOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GoogleAdsUploadOfflineConversionOp(BaseConsumerOp):
    OP_TYPE = "consumer.google.ads.upload_offline_conversion"
    OP_CONFIG_MODEL = GoogleAdsUploadOfflineConversionOpConfigModel
    OP_METADATA_MODEL = GoogleAdsUploadOfflineConversionOpMetadataModel
    OP_AUDIT_MODEL = GoogleAdsUploadOfflineConversionOpAuditModel

    templated_fields = None

    def __init__(
        self,
        google_ads_conn_id: str,
        customer_id: str,
        conversion_action_id: str,
        currency_code: str,
        gclid: str,
        conversion_date_time: str,
        conversion_value: str,
        **kwargs,
    ):
        """."""
        self.google_ads_conn_id = google_ads_conn_id
        self.customer_id = customer_id
        self.conversion_action_id = conversion_action_id
        self.gclid = gclid
        self.conversion_date_time = conversion_date_time
        self.conversion_value = conversion_value
        self.currency_code = currency_code

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.google_ads_conn: GoogleAdsConnectorOp = op_manager.get_connector(
            op_id=self.google_ads_conn_id
        )
        self.client = self.google_ads_conn.client

    def run(self, time_step: TimeStep, msg: Optional[OpMsg] = None) -> OpMsg:
        """."""
        logger.info(f"GoogleAdsUploadOfflineConversionOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="GoogleAdsUploadOfflineConversionOp.run:",
        )

        click_conversion = self.client.get_type("ClickConversion")

        conversion_upload_service = self.client.get_service("ConversionUploadService")
        conversion_action_service = self.client.get_service("ConversionActionService")
        click_conversion.conversion_action = (
            conversion_action_service.conversion_action_path(
                self.customer_id, self.conversion_action_id
            )
        )

        # Sets the single specified ID field.
        click_conversion.gclid = self.gclid

        click_conversion.conversion_value = float(self.conversion_value)
        click_conversion.conversion_date_time = self.conversion_date_time
        click_conversion.currency_code = self.currency_code

        request = self.client.get_type("UploadClickConversionsRequest")
        request.customer_id = self.customer_id
        request.conversions.append(click_conversion)
        request.partial_failure = True
        conversion_upload_response = conversion_upload_service.upload_click_conversions(
            request=request,
        )
        uploaded_click_conversion = conversion_upload_response.results[0]

        if (
            uploaded_click_conversion.conversion_date_time == ""
            and uploaded_click_conversion.gclid == ""
            and uploaded_click_conversion.conversion_action == ""
        ):
            successful_upload = False
            logger.info(
                f"GoogleAdsUploadOfflineConversionOp.run: POSSIBLE FAILED UPLOAD. "
                f"Google Ads API returned an empty response. This most likely means "
                f"that the conversion was not accepted, likely because it was a "
                f"duplicate. The data that was sent to Google Ads is as follows: "
                f"gclid: {self.gclid}, "
                f"conversion_date_time: {self.conversion_date_time}, "
                f"conversion_value: {self.conversion_value}"
            )
        else:
            logger.info("GoogleAdsUploadOfflineConversionOp.run: SUCCESSFUL UPLOAD")
            successful_upload = True

        return OpMsg(
            data=None,
            metadata=GoogleAdsUploadOfflineConversionOpMetadataModel(
                successful_upload=successful_upload,
                gclid=self.gclid,
                conversion_date_time=self.conversion_date_time,
                conversion_value=self.conversion_value,
            ),
            audit=GoogleAdsUploadOfflineConversionOpAuditModel(),
        )



import io
import logging
import pickle
import time
from typing import Generator, Literal, Optional

from pydantic import StrictBool, conint

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class PickleGeneratorOpConfigModel(BaseOpConfigModel):
    output_type: Literal["bytes", "bytesio"]
    protocol: conint(ge=0, le=5) = 4
    fix_imports: Optional[StrictBool] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PickleGeneratorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PickleGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PickleGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.pickle"
    OP_CONFIG_MODEL = PickleGeneratorOpConfigModel
    OP_METADATA_MODEL = PickleGeneratorOpMetadataModel
    OP_AUDIT_MODEL = PickleGeneratorOpAuditModel

    templated_fields = None

    def __init__(
        self,
        output_type: Literal["bytes", "bytesio"],
        protocol: int = 4,
        fix_imports: Optional[StrictBool] = None,
        **kwargs,
    ):
        self.output_type = output_type
        self.protocol = protocol
        self.fix_imports = fix_imports

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"PickleGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="PickleGeneratorOp.run:"
        )

        logger.info(f"PickleGeneratorOp.run: ")

        pickle_io = io.BytesIO()

        pickle_kwargs = {"protocol": self.protocol}
        if self.fix_imports is not None:
            pickle_kwargs["fix_imports"] = self.fix_imports

        pickle.dump(msg.data, pickle_io, **pickle_kwargs)
        pickle_io.seek(0)

        if self.output_type == "bytes":
            output = pickle_io.read()
            f"PickleGeneratorOp.run: pickled bytes {len(output)=:,} bytes."
        elif self.output_type == "bytesio":
            output = pickle_io
        else:
            raise NotImplementedError()

        logger.info(
            f"PickleGeneratorOp.run: Yielding pickled out of type {self.output_type=}."
        )

        yield OpMsg(
            data=output,
            metadata=PickleGeneratorOpMetadataModel(),
            audit=PickleGeneratorOpAuditModel(),
        )



import logging
from io import StringIO
from typing import Any, Dict, Generator, Literal, Optional

import pandas as pd
from pydantic import confloat

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.utils.pandas.log_df_info import log_df_info

logger = logging.getLogger(__name__)


class PandasTrainTestSplitConfigModel(BaseOpConfigModel):
    test_size: confloat(ge=0, le=1)
    random_state: Optional[int] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PandasTrainTestSplitMetadataModel(BaseOpMetadataModel):
    test_size: float
    random_state: Optional[int]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PandasTrainTestSplitAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PandasTrainTestSplit(BaseGeneratorOp):
    OP_TYPE = "generator.pandas.train_and_test_split"
    OP_CONFIG_MODEL = PandasTrainTestSplitConfigModel
    OP_METADATA_MODEL = PandasTrainTestSplitMetadataModel
    OP_AUDIT_MODEL = PandasTrainTestSplitAuditModel

    templated_fields = None

    def __init__(
        self,
        test_size: float,
        random_state: Optional[int],
        **kwargs,
    ):
        self.test_size = test_size
        self.random_state = random_state

        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"PandasTrainTestSplit.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="PandasTrainTestSplit.run:"
        )

        test_df = msg.data.sample(frac=self.test_size, random_state=self.random_state)
        train_df = msg.data.drop(index=test_df.index)

        log_df_info(test_df, log_prefix="PandasTrainTestSplit.run: test_df: ")
        log_df_info(train_df, log_prefix="PandasTrainTestSplit.run: train_df: ")

        yield OpMsg(
            data={
                "train_df": train_df,
                "test_df": test_df,
            },
            metadata=PandasTrainTestSplitMetadataModel(
                test_size=self.test_size,
                random_state=self.random_state,
            ),
            audit=PandasTrainTestSplitAuditModel(),
        )



import logging
from typing import Any, Dict, List, Union

from pydantic import StrictBool, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.postgres import PostgresConnOp

logger = logging.getLogger(__name__)


class SQLQueryGeneratorOpConfigModel(BaseOpConfigModel):
    sql_conn_id: StrictStr
    sql_query: StrictStr
    iterate_over_rows: StrictBool = False

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SQLQueryGeneratorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SQLQueryGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class SQLQueryGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.sql_query"
    OP_CONFIG_MODEL = SQLQueryGeneratorOpConfigModel
    OP_METADATA_MODEL = SQLQueryGeneratorOpMetadataModel
    OP_AUDIT_MODEL = SQLQueryGeneratorOpAuditModel

    templated_fields = None

    def __init__(
        self,
        sql_conn_id: str,
        sql_query: str,
        iterate_over_rows: bool = False,
        **kwargs,
    ):
        self.sql_conn_id = sql_conn_id
        self.sql_query = sql_query
        self.iterate_over_rows = iterate_over_rows
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.db: Union[PostgresConnOp] = op_manager.get_connector(
            op_id=self.sql_conn_id
        )

    def run(self, time_step: TimeStep, msg: OpMsg) -> OpMsg:
        """."""
        logger.info(f"SQLQueryGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="SQLQueryGeneratorOp.run:"
        )

        with self.db.get_engine().connect() as conn:
            result = conn.execute(self.sql_query)

        if self.iterate_over_rows:
            for row in result:
                yield OpMsg(
                    data=row,
                    metadata=SQLQueryGeneratorOpMetadataModel(),
                    audit=SQLQueryGeneratorOpAuditModel(),
                )
        else:
            yield OpMsg(
                data=result,
                metadata=SQLQueryGeneratorOpMetadataModel(),
                audit=SQLQueryGeneratorOpAuditModel(),
            )



import bz2
import gzip
import logging
import zlib
from typing import Any, Dict, Generator, List, Literal, Optional

import zstandard as zstd
from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class DecompressOpConfigModel(BaseOpConfigModel):
    decompress_type: Literal["bz2", "gzip", "zlib", "zstd"]
    decompress_kwargs: Optional[Dict[StrictStr, Any]] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DecompressOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DecompressOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DecompressOp(BaseGeneratorOp):
    OP_TYPE = "generator.decompress"
    OP_CONFIG_MODEL = DecompressOpConfigModel
    OP_METADATA_MODEL = DecompressOpMetadataModel
    OP_AUDIT_MODEL = DecompressOpAuditModel

    templated_fields = None

    def __init__(
        self,
        decompress_type: str,
        decompress_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.decompress_type = decompress_type
        self.decompress_kwargs = {} if decompress_kwargs is None else decompress_kwargs
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"DecompressOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=time_step, msg=msg, log_prefix="DecompressOp.run:")

        if self.decompress_type == "bz2":
            decompress = bz2.decompress
        elif self.decompress_type == "gzip":
            decompress = gzip.decompress
        elif self.decompress_type == "zlib":
            decompress = zlib.decompress
        elif self.decompress_type == "zstd":
            decompress = zstd.decompress
        else:
            raise NotImplementedError()

        if not isinstance(msg.data, bytes):
            raise ValueError(
                f"DecompressOp.run: msg.data must be bytes, not {type(msg.data)}"
            )

        output = decompress(msg.data, **self.decompress_kwargs)

        logger.info(
            f"DecompressOp.run: Decompressed {len(msg.data):,} bytes "
            f"to {len(output):,} bytes. "
            f"Decompression ratio: {len(output) / len(msg.data):.2f}"
        )

        yield OpMsg(
            data=output,
            metadata=DecompressOpMetadataModel(),
            audit=DecompressOpAuditModel(),
        )



import logging
from typing import Generator, Literal, Optional, Union

from pydantic import StrictBool, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.list import ListConnOp

logger = logging.getLogger(__name__)


class ListGetGeneratorOpConfigModel(BaseOpConfigModel):
    list_conn_id: StrictStr
    get_type: Literal["iterate_over_list", "get_reference"]
    reinitialize_list: StrictBool

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ListGetGeneratorOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ListGetGeneratorOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class ListGetGeneratorOp(BaseGeneratorOp):
    OP_TYPE = "generator.list.get"
    OP_CONFIG_MODEL = ListGetGeneratorOpConfigModel
    OP_METADATA_MODEL = ListGetGeneratorOpMetadataModel
    OP_AUDIT_MODEL = ListGetGeneratorOpAuditModel

    templated_fields = None

    def __init__(
        self,
        list_conn_id: str,
        get_type: str,
        reinitialize_list: bool,
        **kwargs,
    ):
        self.list_conn_id = list_conn_id
        self.get_type = get_type
        self.reinitialize_list = reinitialize_list
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.list_conn: Union[ListConnOp] = op_manager.get_connector(
            op_id=self.list_conn_id
        )

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"ListGetGeneratorOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="ListGetGeneratorOp.run:"
        )

        if self.reinitialize_list:
            _list = self.list_conn.get_reference_and_reinit()
        else:
            _list = self.list_conn.get_reference()

        if self.get_type == "get_reference":
            logger.info(
                f"ListGetGeneratorOp.run: Yielding reference to list. {len(_list)=}"
            )
            yield OpMsg(
                data=_list,
                metadata=ListGetGeneratorOpMetadataModel(),
                audit=ListGetGeneratorOpAuditModel(),
            )

        elif self.get_type == "iterate_over_list":
            logger.info(
                f"ListGetGeneratorOp.run: Iterating over list and yielding items. "
                f"{len(_list)=}"
            )
            for item_i, item in enumerate(_list):
                logger.info(
                    f"ListGetGeneratorOp.run: Yielding {item_i=} (0-index) from list. "
                    f"{type(item)=}"
                )
                yield OpMsg(
                    data=item,
                    metadata=ListGetGeneratorOpMetadataModel(),
                    audit=ListGetGeneratorOpAuditModel(),
                )
        else:
            raise NotImplementedError()

        if self.reinitialize_list:
            logger.info(
                f"ListGetGeneratorOp.run: Done yielding. {self.reinitialize_list=}. "
                f"List has been reinitialized to an empty list []."
            )
        else:
            logger.info(
                f"ListGetGeneratorOp.run: Done yielding. {self.reinitialize_list=}. "
                f"List has not been reinitialized."
            )



import json
import logging
import time
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union

import pandas as pd
from pydantic import StrictBool, StrictInt, StrictStr, conlist, root_validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.http import HTTPConnOp

logger = logging.getLogger(__name__)


class OngoingOrdersRESTGetOrderOpConfigModel(BaseOpConfigModel):
    http_conn_id: StrictStr
    accepted_status_codes: Optional[conlist(int, min_items=1)] = None
    timeout_seconds_per_request: float = 60
    order_id: Optional[StrictStr] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingOrdersRESTGetOrderOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingOrdersRESTGetOrderOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class OngoingOrdersRESTGetOrderOp(BaseGeneratorOp):
    """."""

    OP_TYPE = "generator.ongoing.orders.rest_get_order"
    OP_CONFIG_MODEL = OngoingOrdersRESTGetOrderOpConfigModel
    OP_METADATA_MODEL = OngoingOrdersRESTGetOrderOpMetadataModel
    OP_AUDIT_MODEL = OngoingOrdersRESTGetOrderOpAuditModel

    templated_fields = None

    def __init__(
        self,
        http_conn_id: str,
        order_id: Optional[str] = None,
        accepted_status_codes: Optional[List[int]] = None,
        timeout_seconds_per_request: float = 60,
        **kwargs,
    ):
        """."""
        self.http_conn_id = http_conn_id
        self.order_id = order_id
        self.accepted_status_codes = (
            accepted_status_codes if accepted_status_codes else [200]
        )
        self.timeout_seconds_per_request = timeout_seconds_per_request

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.http_conn: HTTPConnOp = op_manager.get_connector(op_id=self.http_conn_id)

    def get_page(self, order_id: str) -> Tuple[int, Dict[str, Any], Dict[str, Any]]:
        """."""
        params = {}

        # DOCS: https://developer.ongoingwarehouse.com/REST/v1/index.html#/Orders/Orders_Get
        url = f"{self.http_conn.base_url}/api/v1/orders/{order_id}"
        logger.info(
            f"OngoingOrdersRESTGetOrderOp.get_page: Fetching order for "
            f"{order_id=} with {url=}"
        )

        etl_request_start_ts = pd.Timestamp.now(tz="UTC").isoformat()
        status_code, output_json = self.http_conn.call(
            method="get",
            url=url,
            requests_kwargs={
                "params": params,
                "timeout": self.timeout_seconds_per_request,
            },
            # NOTE: We expect status code to be 200 with return type JSON
            # even if there are no records.
            accepted_status_codes=self.accepted_status_codes,
            return_type="json",
        )
        etl_response_end_ts = pd.Timestamp.now(tz="UTC").isoformat()

        etl_metadata_json = {
            "etl_request_start_ts": etl_request_start_ts,
            "etl_response_end_ts": etl_response_end_ts,
        }

        return status_code, output_json, etl_metadata_json

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"OngoingOrdersRESTGetOrderOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="OngoingOrdersRESTGetOrderOp.run:",
        )

        if self.order_id:
            order_ids = [self.order_id]
        else:
            raise NotImplementedError()

        final_records = []
        for order_id in order_ids:
            logger.info(
                f"OngoingOrdersRESTGetOrderOp.run: Fetching order for {order_id=}."
            )

            status_code, output_json, etl_metadata_json = self.get_page(
                order_id=order_id
            )

            logger.info(
                f"OngoingOrdersRESTGetOrderOp.run: Done fetching order "
                f"for {order_id=}."
            )

            ############################################################################
            # NOTE: Adding `etl_metadata` into response json (output_json).
            # * This is to add some metadata about the ETL process, primarily some
            #   timestamps which may help with creating event stores/logs.
            # * This may also help with incremental modelling.
            # * DO NOT USE this information when calculating dedup_uuid.
            ############################################################################
            etl_metadata_json["time_step"] = time_step.to_json_dict()
            records = [
                {
                    "data": output_json,
                    "etl_metadata": etl_metadata_json,
                }
            ]

            final_records.extend(records)

        yield OpMsg(
            data=final_records,
            metadata=OngoingOrdersRESTGetOrderOpMetadataModel(),
            audit=OngoingOrdersRESTGetOrderOpAuditModel(),
        )



import logging
from typing import Dict, Generator, List, Optional

from pydantic import StrictStr, conlist

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.google_analytics_ga4 import (
    GoogleAnalyticsGA4ConnOp,
)

logger = logging.getLogger(__name__)


# Reference: https://developers.google.com/analytics/devguides/reporting/core/v4/quickstart/service-py
class ReportConfig(BaseOpConfigModel):
    report_name: StrictStr
    report_docs: Optional[StrictStr] = None

    # Keeping the type enforcements to a minimum since I don't fully understand the API
    property_id: StrictStr
    date_ranges: Optional[List[Dict[str, str]]] = None
    metrics: Optional[List[Dict[str, str]]] = None
    dimensions: Optional[List[Dict[str, str]]] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GetGAReportsGA4OpConfigModel(BaseOpConfigModel):
    google_analytics_conn_id: StrictStr
    report_configs: conlist(ReportConfig, min_items=1)

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GetGAReportsGA4OpMetadataModel(BaseOpMetadataModel):
    property_id: StrictStr
    report_name: StrictStr
    report_docs: Optional[StrictStr] = None

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GetGAReportsGA4OpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GetGAReportsGA4Op(BaseGeneratorOp):
    OP_TYPE = "generator.get_ga_reports_ga4"
    OP_CONFIG_MODEL = GetGAReportsGA4OpConfigModel
    OP_METADATA_MODEL = GetGAReportsGA4OpMetadataModel
    OP_AUDIT_MODEL = GetGAReportsGA4OpAuditModel

    templated_fields = None

    def __init__(
        self,
        google_analytics_conn_id: str,
        report_configs: List[Dict[str, str]],
        **kwargs,
    ):
        """."""
        self.google_analytics_conn_id = google_analytics_conn_id
        self.report_configs = report_configs
        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.analytics: GoogleAnalyticsGA4ConnOp = op_manager.get_connector(
            op_id=self.google_analytics_conn_id
        )

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"GetGAReportsGA4Op.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="GetGAReportsGA4Op.run:"
        )

        for report_config in self.report_configs:
            logger.info(
                f"GetGAReportsGA4Op.run: Fetching report: "
                f"{report_config['report_name']}"
            )
            report = self.analytics.get_report(**report_config)

            logger.info(
                f"GetGAReportsGA4Op.run: Fetching report: "
                f"{report_config['report_name']} Done! {type(report)=}"
            )

            yield OpMsg(
                data=report,
                metadata=GetGAReportsGA4OpMetadataModel(
                    property_id=report_config.get("property_id"),
                    report_name=report_config.get("report_name"),
                    report_docs=report_config.get("report_docs"),
                ),
                audit=GetGAReportsGA4OpAuditModel(),
            )



"""."""
import json
import logging
import os
from typing import Any, Dict, Literal, Optional

from pydantic import BaseModel, Field, StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class HTTPRequestsHookFromEnvConnOpConfigModel(BaseOpConfigModel):
    env_var: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class HTTPRequestsHookFromEnvConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class HTTPRequestsHookFromEnvConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class HTTPConnEnvModel(BaseModel):
    conn_type: Literal["http"]
    host: StrictStr
    login: Optional[StrictStr] = None
    password: Optional[StrictStr] = None
    extras: Dict[str, Any] = Field(default_factory=dict)

    class Config:
        extra = "allow"
        arbitrary_types_allowed = False


class HTTPRequestsHookFromEnvConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.http_requests_from_env"
    OP_CONFIG_MODEL = HTTPRequestsHookFromEnvConnOpConfigModel
    OP_METADATA_MODEL = HTTPRequestsHookFromEnvConnOpMetadataModel
    OP_AUDIT_MODEL = HTTPRequestsHookFromEnvConnOpAuditModel

    templated_fields = None

    def __init__(self, env_var: str, **kwargs):
        self.env_var = env_var
        self.templated_fields = kwargs.get("templated_fields", [])

        # Loaded in run
        self.conn_model: Optional[HTTPConnEnvModel] = None

    def run(self) -> OpMsg:
        """."""
        logger.info(f"HTTPRequestsHookFromEnvConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="HTTPRequestsHookFromEnvConnOp.run:"
        )

        if self.env_var not in os.environ:
            raise ValueError(
                f"HTTPRequestsHookFromEnvConnOp.run: Environment variable "
                f"{self.env_var} not found!"
            )

        conn_json = json.loads(os.environ[self.env_var])
        self.conn_model = HTTPConnEnvModel(**conn_json)

        return OpMsg(
            data=self,
            metadata=HTTPRequestsHookFromEnvConnOpMetadataModel(),
            audit=HTTPRequestsHookFromEnvConnOpAuditModel(),
        )

    ####################################################################################
    # PROPERTIES
    ####################################################################################
    @property
    def base_url(self) -> str:
        """."""
        _base_url = self.conn_model.host
        if _base_url.endswith("/"):
            _base_url = self.conn_model.host[:-1]
        return _base_url

    @property
    def username(self) -> Optional[str]:
        """."""
        return self.conn_model.login

    @property
    def password(self) -> Optional[str]:
        """."""
        return self.conn_model.password

    @property
    def extras(self) -> dict:
        """."""
        return self.conn_model.extras



import logging
from typing import Any, Dict, List, Literal, Optional

from pydantic import StrictBool, StrictStr, validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.s3 import S3ConnOp

logger = logging.getLogger(__name__)


class S3PutKeyOpConfigModel(BaseOpConfigModel):
    s3_conn_id: StrictStr
    bucket: StrictStr
    key: StrictStr
    replace: StrictBool
    input_type: Literal["bytes", "bytesio"]

    @validator("key", pre=False)
    def validate_key(cls, key: str):
        if key.endswith("/"):
            raise ValueError(f"{key=} must not end with a '/'")

        if key.startswith("/"):
            raise ValueError(f"{key=} must not start with a '/'")

        return key

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class S3PutKeyOpMetadataModel(BaseOpMetadataModel):
    bucket: StrictStr
    key: StrictStr
    input_type: Literal["bytes", "bytesio"]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class S3PutKeyOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class S3PutKeyOp(BaseConsumerOp):
    OP_TYPE = "consumer.s3.put_key"
    OP_CONFIG_MODEL = S3PutKeyOpConfigModel
    OP_METADATA_MODEL = S3PutKeyOpMetadataModel
    OP_AUDIT_MODEL = S3PutKeyOpAuditModel

    templated_fields = None

    def __init__(
        self,
        s3_conn_id: str,
        bucket: str,
        key: str,
        replace: bool,
        input_type: Literal["bytes", "bytesio"],
        templated_fields: List[Dict[str, Any]] = None,
        **kwargs,
    ):
        """."""
        self.s3_conn_id = s3_conn_id
        self.bucket = bucket
        self.key = key
        self.replace = replace
        self.input_type = input_type
        self.templated_fields = [] if templated_fields is None else templated_fields

        op_manager = get_global_op_manager()

        self.s3_conn: S3ConnOp = op_manager.get_connector(op_id=self.s3_conn_id)

    def run(self, time_step: TimeStep, msg: Optional[OpMsg] = None) -> OpMsg:
        """."""
        logger.info(f"S3PutKeyOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(time_step=time_step, msg=None, log_prefix="S3PutKeyOp.run:")

        logger.info(
            f"S3PutKeyOp.run: Putting key {self.key=} with {self.input_type=}. "
            f"Input {type(msg.data)=}"
        )

        if self.input_type == "bytes":
            logger.info(
                f"S3PutKeyOp.run: Uploading with {len(msg.data):,} bytes "
                f"i.e. approx {len(msg.data)//1000//1000:,} MiB"
            )

        self.s3_conn.put_key(
            bucket=self.bucket,
            key=self.key,
            input_type=self.input_type,
            replace=self.replace,
            data=msg.data,
        )

        logger.info(
            f"S3PutKeyOp.run: Uploading object to s3 "
            f"{self.input_type=} | {type(msg.data)=} | to {self.bucket} | {self.key=}"
        )

        return OpMsg(
            data=None,
            metadata=S3PutKeyOpMetadataModel(
                bucket=self.bucket,
                key=self.key,
                input_type=self.input_type,
            ),
            audit=S3PutKeyOpAuditModel(),
        )



"""."""
import logging

from pydantic import StrictStr

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class AirflowS3HookConnOpConfigModel(BaseOpConfigModel):
    airflow_conn_id: StrictStr

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowS3HookConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowS3HookConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class AirflowS3HookConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.airflow_s3_hook"
    OP_CONFIG_MODEL = AirflowS3HookConnOpConfigModel
    OP_METADATA_MODEL = AirflowS3HookConnOpMetadataModel
    OP_AUDIT_MODEL = AirflowS3HookConnOpAuditModel

    templated_fields = None

    def __init__(self, airflow_conn_id: str, **kwargs):
        self.airflow_conn_id = airflow_conn_id
        self.templated_fields = kwargs.get("templated_fields", [])

        from airflow.providers.amazon.aws.hooks.s3 import S3Hook

        self.hook = S3Hook(aws_conn_id=self.airflow_conn_id)

    def run(self) -> OpMsg:
        """."""
        logger.info(f"AirflowS3HookConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="AirflowS3HookConnOp.run:"
        )

        return OpMsg(
            data=self.hook,
            metadata=AirflowS3HookConnOpMetadataModel(),
            audit=AirflowS3HookConnOpAuditModel(),
        )



import bz2
import gzip
import itertools
import json
import logging
import zlib
from typing import Any, Dict, Generator, List, Literal, Optional

import zstandard as zstd
from pydantic import StrictStr, StrictInt

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseGeneratorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class DecompressNewlineSeparatedStreamsWithZlibOpConfigModel(BaseOpConfigModel):
    wbits: StrictInt

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DecompressNewlineSeparatedStreamsWithZlibOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DecompressNewlineSeparatedStreamsWithZlibOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class DecompressNewlineSeparatedStreamsWithZlibOp(BaseGeneratorOp):
    OP_TYPE = "generator.decompress.decompress_newline_separated_streams_with_zlib"
    OP_CONFIG_MODEL = DecompressNewlineSeparatedStreamsWithZlibOpConfigModel
    OP_METADATA_MODEL = DecompressNewlineSeparatedStreamsWithZlibOpMetadataModel
    OP_AUDIT_MODEL = DecompressNewlineSeparatedStreamsWithZlibOpAuditModel

    templated_fields = None

    def __init__(
        self,
        wbits: int,
        **kwargs,
    ):
        self.wbits = wbits
        self.templated_fields = kwargs.get("templated_fields", [])

    def run(
        self, time_step: TimeStep, msg: Optional[OpMsg] = None
    ) -> Generator[OpMsg, None, None]:
        """."""
        logger.info(f"DecompressNewlineSeparatedStreamsWithZlibOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step,
            msg=msg,
            log_prefix="DecompressNewlineSeparatedStreamsWithZlibOp.run:",
        )

        current_data = msg.data

        for iter_i in itertools.count(0):
            try:
                z = zlib.decompressobj(wbits=47)
                output = z.decompress(current_data)
                unused_data = z.unused_data
                logger.info(
                    f"DecompressNewlineSeparatedStreamsWithZlibOp.run: "
                    f"Done decompressing {iter_i=}. "
                    f"{len(output)=:,} bytes. "
                    f"{len(unused_data)=:,} bytes. Yielding uncompressed output."
                )
                yield OpMsg(
                    data=output,
                    metadata=DecompressNewlineSeparatedStreamsWithZlibOpMetadataModel(),
                    audit=DecompressNewlineSeparatedStreamsWithZlibOpAuditModel(),
                )

            except Exception:
                logger.error(
                    f"DecompressNewlineSeparatedStreamsWithZlibOp.run: We do not expect "
                    f"any errors in this operation. Please investigate and make sure "
                    f"the input data provided fits the use case that this op tries to "
                    f"solve."
                )
                raise

            if unused_data.startswith(b"\n"):
                unused_data = unused_data[1:]

            if len(unused_data) == 0:
                break

            current_data = unused_data

        logger.info(
            f"DecompressNewlineSeparatedStreamsWithZlibOp.run: "
            f"No more unused_data to process. Returning. "
        )



"""."""
import json
import logging
from collections import Counter
from typing import Dict, List

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.ops.base import BaseConnectorOp, BaseOpConfigModel

logger = logging.getLogger(__name__)


class PythonCounterHookConnOpConfigModel(BaseOpConfigModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonCounterHookConnOpMetadataModel(BaseOpMetadataModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonCounterHookConnOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class PythonCounterHookConnOp(BaseConnectorOp):
    OP_TYPE = "connector.hooks.python.counter"
    OP_CONFIG_MODEL = PythonCounterHookConnOpConfigModel
    OP_METADATA_MODEL = PythonCounterHookConnOpMetadataModel
    OP_AUDIT_MODEL = PythonCounterHookConnOpAuditModel

    templated_fields = None

    def __init__(self, **kwargs):
        self.templated_fields = kwargs.get("templated_fields", [])

        # NOTE: This is the counter :)
        self._counter = Counter()

    def run(self) -> OpMsg:
        """."""
        logger.info(f"PythonCounterHookConnOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=None, msg=None, log_prefix="PythonCounterHookConnOp.run:"
        )

        return OpMsg(
            data=self,
            metadata=PythonCounterHookConnOpMetadataModel(),
            audit=PythonCounterHookConnOpAuditModel(),
        )

    def get_reference(self):
        """."""
        return self._counter

    def get_reference_and_reinit(self):
        """."""
        _counter = self._counter
        self._counter = Counter()
        return _counter

    def put_update_dict(self, item: Dict[str, int]):
        """."""
        self._counter.update(item)
        logger.info(
            f"PythonCounterHookConnOp.put_update_dict: Counter updated. "
            f"New counter value: {json.dumps(self._counter, indent=2)}"
        )

    def put_update_list_of_dicts(self, item: List[Dict[str, int]]):
        """."""
        for _item in item:
            self._counter.update(_item)
        logger.info(
            f"PythonCounterHookConnOp.put_update_list_of_dicts: Counter updated. "
            f"New counter value: {json.dumps(self._counter, indent=2)}"
        )



import logging
from typing import Literal, Optional

from pydantic import StrictBool, StrictStr, validator

from nr_ops.messages.op_audit import BaseOpAuditModel
from nr_ops.messages.op_metadata import BaseOpMetadataModel
from nr_ops.messages.op_msg import OpMsg
from nr_ops.messages.time_step import TimeStep
from nr_ops.ops.base import BaseConsumerOp, BaseOpConfigModel
from nr_ops.ops.op_manager import get_global_op_manager
from nr_ops.ops.ops.connector_ops.interfaces.gcp_gcs import GCPGCSConnOp

logger = logging.getLogger(__name__)


class GCPGCSPutKeyOpConfigModel(BaseOpConfigModel):
    gcs_conn_id: StrictStr
    bucket: StrictStr
    key: StrictStr
    replace: StrictBool
    input_type: Literal["bytes", "bytesio"]

    @validator("key", pre=False)
    def validate_key(cls, key: str):
        if key.endswith("/"):
            raise ValueError(f"{key=} must not end with a '/'")

        if key.startswith("/"):
            raise ValueError(f"{key=} must not start with a '/'")

        return key

    @validator("replace", pre=False)
    def validate_replace(cls, replace: bool):
        if replace is True:
            raise ValueError(
                f"{replace=} is not implemented. Must be False. If you require "
                f"replace functionality, then use `gcs.is_key_exists` and "
                f"`gcs.delete_key` operators to implement your own."
            )

        return replace

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSPutKeyOpMetadataModel(BaseOpMetadataModel):
    bucket: StrictStr
    key: StrictStr
    input_type: Literal["bytes", "bytesio"]

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSPutKeyOpAuditModel(BaseOpAuditModel):
    pass

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = False


class GCPGCSPutKeyOp(BaseConsumerOp):
    OP_TYPE = "consumer.gcp.gcs.put_key"
    OP_CONFIG_MODEL = GCPGCSPutKeyOpConfigModel
    OP_METADATA_MODEL = GCPGCSPutKeyOpMetadataModel
    OP_AUDIT_MODEL = GCPGCSPutKeyOpAuditModel

    templated_fields = None

    def __init__(
        self,
        gcs_conn_id: str,
        bucket: str,
        key: str,
        replace: bool,
        input_type: Literal["bytes", "bytesio"],
        **kwargs,
    ):
        """."""
        self.gcs_conn_id = gcs_conn_id
        self.bucket = bucket
        self.key = key
        self.replace = replace
        self.input_type = input_type

        self.templated_fields = kwargs.get("templated_fields", [])

        op_manager = get_global_op_manager()

        self.gcs_conn: GCPGCSConnOp = op_manager.get_connector(op_id=self.gcs_conn_id)

    def run(self, time_step: TimeStep, msg: Optional[OpMsg] = None) -> OpMsg:
        """."""
        logger.info(f"GCPGCSPutKeyOp.run: Running")

        # RENDERS AND UPDATES THE TEMPLATED FIELDS INPLACE
        self.render_fields(
            time_step=time_step, msg=msg, log_prefix="GCPGCSPutKeyOp.run:"
        )

        logger.info(
            f"GCPGCSPutKeyOp.run: Uploading object to GCS "
            f"{self.input_type=} | {type(msg.data)=} | to {self.bucket} | {self.key=}"
        )

        self.gcs_conn.put_key(
            bucket=self.bucket,
            key=self.key,
            data=msg.data,
            input_type=self.input_type,
            replace=self.replace,
        )

        logger.info(
            f"GCPGCSPutKeyOp.run: Done uploading to GCS with {self.input_type=} for "
            f"{self.key=}"
        )
        return OpMsg(
            data=None,
            metadata=GCPGCSPutKeyOpMetadataModel(
                bucket=self.bucket,
                key=self.key,
                input_type=self.input_type,
            ),
            audit=GCPGCSPutKeyOpAuditModel(),
        )
